#include "trayiconwatcher.h"
#include "utils/misc.h"
#include <QDBusReply>

TrayIconWatcher::TrayIconWatcher(QObject *parent) : QObject(parent)
{
}

QString TrayIconWatcher::iconSource()
{
    return m_iconSource;
}

void TrayIconWatcher::setIconSource(const QString &source)
{
    m_iconSource = source;
    emit iconSourceChanged();
}

int TrayIconWatcher::type()
{
    return m_type;
}

void TrayIconWatcher::setType(const int &type)
{
    m_type = type;
    switch (m_type) {
    case Volume: {
        QDBusConnection::sessionBus().connect(QString("org.ukui.media"),
                                              QString("/org/ukui/media"),
                                              QString("org.ukui.media"),
                                              QString("updateMute"), this, SLOT(setMuteState(bool)));
        QDBusConnection::sessionBus().connect(QString("org.ukui.media"),
                                              QString("/org/ukui/media"),
                                              QString("org.ukui.media"),
                                              QString("updateVolume"), this, SLOT(setVolume(int)));
        getVolume();
        break;
    }
    case Power: {
        initPowerDbusaddrest();
        initPowerValue();
        QDBusConnection::systemBus().connect(QString("org.freedesktop.UPower"),
                                             powerpath,
                                             "org.freedesktop.DBus.Properties",
                                             "PropertiesChanged",
                                             this,
                                             SLOT(initPowerValue()));
        initPowerValue();
        break;
    }
    case Message: {
        QDBusConnection::sessionBus().connect(QString("org.ukui.Sidebar"),
                                              QString("/org/ukui/Sidebar/notification"),
                                              QString("org.ukui.Sidebar.notification"),
                                              QString("Sig_NotificationNumChange"),
                                              this,
                                              SLOT(getMessageState()));

        getMessageState();
        break;
    }
    case Network: {
        getWifiValue();
        QDBusConnection::systemBus().connect(QString("org.freedesktop.NetworkManager"),
                                             QString("/org/freedesktop/NetworkManager"),
                                             QString("org.freedesktop.NetworkManager"),
                                             QString("PropertiesChanged"),
                                             this,
                                             SLOT(getWifiValue()));
        getWifiValue();
        break;
    }
    case Bluetooth: {
        QDBusConnection::sessionBus().connect(QString("com.ukui.bluetooth"),
                                              QString("/com/ukui/bluetooth"),
                                              QString("com.ukui.bluetooth"),
                                              QString("defaultAdapterPowerChanged"),
                                              this,
                                              SLOT(getBluetoothSwitch()));

        getBluetoothSwitch();
        break;
    }
    case TimeDate: {
        if (QGSettings::isSchemaInstalled("org.ukui.control-center.panel.plugins")) {
            timeSetting = new QGSettings("org.ukui.control-center.panel.plugins");
            connect(timeSetting, &QGSettings::changed, this, &TrayIconWatcher::getTimeDateState);
        }
        getTimeDateState();
    }
    default:
        break;
    }
    emit typeChanged();
}

QString TrayIconWatcher::timeformat()
{
    return m_timeformat;
}

void TrayIconWatcher::settimeformat(const QString &timeformat)
{
    if (m_timeformat != timeformat) {
        m_timeformat = timeformat;
        emit timeformatChanged();
    }
}

QString TrayIconWatcher::dateformat()
{
    return m_dateformat;
}

void TrayIconWatcher::setdateformat(const QString &dateformat)
{
    if (m_dateformat != dateformat) {
        m_dateformat = dateformat;
        emit dateformatChanged();
    }
}

void TrayIconWatcher::getVolume()
{
    auto reply = Misc::DbusAsyncCall<int>("org.ukui.media",
                                          "/org/ukui/media",
                                          "org.ukui.media",
                                          QDBusConnection::sessionBus(),
                                          "getDefaultOutputVolume");
    if (reply.isError()) {
        qWarning() << "getDefaultOutputVolume error, " << reply.error();
        setVolumeIcon(0);
        return;
    }
    setVolumeIcon(reply.value());
}

void TrayIconWatcher::initPowerDbusaddrest()
{
    auto reply = Misc::DbusAsyncCall<QList<QDBusObjectPath>>(
                "org.freedesktop.UPower",
                 "/org/freedesktop/UPower",
                 "org.freedesktop.UPower",
                 QDBusConnection::systemBus(),
                 "EnumerateDevices");

    if (reply.isError()) {
        qWarning() << "EnumerateDevices error, " << reply.error();
        return;
    }

    QStringList users;
    for (QDBusObjectPath op : reply.value()) {
         users << op.path();
    }
    if (users.count() <= 1) {
        return;
    }
    powerpath = users.at(1);
}

void TrayIconWatcher::getMessageState()
{
    auto reply = Misc::DbusAsyncCall<int>("org.ukui.Sidebar",
                                          "/org/ukui/Sidebar/notification",
                                          "org.ukui.Sidebar.notification",
                                          QDBusConnection::sessionBus(),
                                          "getNotificationNum");
    if (reply.isError()) {
        qWarning() << "getNotificationNum error, " << reply.error();
        setMessageIcon(false);
        return;
    }
    setMessageIcon(reply.value());
}

void TrayIconWatcher::initPowerValue()
{
    auto percentageReply = Misc::DbusAsyncCallWithArgs<QVariant>("org.freedesktop.UPower",
                                                                 powerpath,
                                                                 "org.freedesktop.DBus.Properties",
                                                                 QDBusConnection::systemBus(),
                                                                 "Get",
                                                                 QList<QVariant>()
                                                                     << "org.freedesktop.UPower.Device"
                                                                     << "Percentage");
    if (percentageReply.isError()) {
        qWarning() << "Get org.freedesktop.UPower.Device.Percentage error, " << percentageReply.error();
    }

    auto stateReply = Misc::DbusAsyncCallWithArgs<QVariant>("org.freedesktop.UPower",
                                                            powerpath,
                                                            "org.freedesktop.DBus.Properties",
                                                            QDBusConnection::systemBus(),
                                                            "Get",
                                                            QList<QVariant>()
                                                                << "org.freedesktop.UPower.Device"
                                                                << "State");

    if (stateReply.isError()) {
        qWarning() << "Get org.freedesktop.UPower.Device.State error, " << stateReply.error();
    }

    if (percentageReply.isValid() && stateReply.isValid()) {
        setPowerIcon(stateReply.value().toInt(), percentageReply.value().toInt());
    }
}

void TrayIconWatcher::getBluetoothSwitch()
{
    auto adapterAddressReply = Misc::DbusAsyncCall<QString>("com.ukui.bluetooth",
                                                            "/com/ukui/bluetooth",
                                                            "com.ukui.bluetooth",
                                                            QDBusConnection::sessionBus(),
                                                            "getDefaultAdapterAddress");
    if (adapterAddressReply.isError()) {
        qWarning() << "getDefaultAdapterAddress error, " << adapterAddressReply.error();
        setBluetoothIcon(false);
        return;
    }

    if (adapterAddressReply.value().isEmpty()) {
        setBluetoothIcon(true);
        return;
    }

    auto adapterPowerReply = Misc::DbusAsyncCall<bool>("com.ukui.bluetooth",
                                                       "/com/ukui/bluetooth",
                                                       "com.ukui.bluetooth",
                                                       QDBusConnection::sessionBus(),
                                                       "getDefaultAdapterPower");
    if (adapterPowerReply.isError()) {
        qWarning() << "getDefaultAdapterPower error, " << adapterAddressReply.error();
        setBluetoothIcon(false);
        return;
    }

    setBluetoothIcon(adapterPowerReply.value());
}

void TrayIconWatcher::getTimeDateState()
{
    if (QGSettings::isSchemaInstalled("org.ukui.control-center.panel.plugins")) {
        if (timeSetting->keys().contains("hoursystem")) {
            QString timeState = timeSetting->get("hoursystem").toString();
            settimeformat(timeState);
        }
        if (timeSetting->keys().contains("hoursystem")) {
            QString dateState = timeSetting->get("date").toString();
            setdateformat(dateState);
        }
    }
}


void TrayIconWatcher::getWifiValue()
{
    netconnectType = 0;

    QDBusInterface interface( "org.freedesktop.NetworkManager",
                              "/org/freedesktop/NetworkManager",
                              "org.freedesktop.DBus.Properties",
                              QDBusConnection::systemBus());

    //获取已经连接了那些网络，及这些网络对应的网络类型(ethernet or wifi)
    QDBusMessage result = interface.call("Get", "org.freedesktop.NetworkManager", "ActiveConnections");
    QList<QVariant> outArgs = result.arguments();
    QVariant first = outArgs.at(0);
    QDBusVariant dbvFirst = first.value<QDBusVariant>();
    QVariant vFirst = dbvFirst.variant();
    QDBusArgument dbusArgs = vFirst.value<QDBusArgument>();
    QDBusArgument dbusArgshelp = vFirst.value<QDBusArgument>();

    QDBusObjectPath objPath;
    dbusArgs.beginArray();

    while (!dbusArgs.atEnd()) {
        dbusArgs >> objPath;
        QDBusInterface interfacePath( "org.freedesktop.NetworkManager",
                                  objPath.path(),
                                  "org.freedesktop.DBus.Properties",
                                  QDBusConnection::systemBus() );

        QDBusReply<QVariant> replyConnectType = interfacePath.call("Get", "org.freedesktop.NetworkManager.Connection.Active", "Type");
        QDBusReply<QVariant> replyConnectivity = interface.call("Get", "org.freedesktop.NetworkManager", "Connectivity");
        int connectivityState = replyConnectivity.value().toInt();
        if (replyConnectType.value().toString() == "802-3-ethernet") {
            netconnectType = 1;
            if (connectivityState == 4) {
                setNetIcon(-1, true); //有网
                return;
            } else {
                setNetIcon(-2, true); //无网
            }
        }

        if (replyConnectType.value().toString() == "802-11-wireless") {
            netconnectType = 2;
            QDBusReply<QVariant> replyConnectivity = interface.call("Get", "org.freedesktop.NetworkManager", "Connectivity");
            if (connectivityState == 4) {
                QDBusMessage resultSpecificObject = interfacePath.call("Get", "org.freedesktop.NetworkManager.Connection.Active", "SpecificObject");
                QList<QVariant> outArgsSpecificObject = resultSpecificObject.arguments();
                QVariant firstSpecificObject = outArgsSpecificObject.at(0);
                QDBusVariant dbvFirstSpecificObject = firstSpecificObject.value<QDBusVariant>();
                QVariant vFirstSpecificObject = dbvFirstSpecificObject.variant();
                QDBusObjectPath objPathSpecificObject = vFirstSpecificObject.value<QDBusObjectPath>();

                QDBusInterface interfaceAP( "org.freedesktop.NetworkManager",
                                          objPathSpecificObject.path(),
                                          "org.freedesktop.DBus.Properties",
                                          QDBusConnection::systemBus() );

                QDBusReply<QVariant> replyStrength = interfaceAP.call("Get", "org.freedesktop.NetworkManager.AccessPoint", "Strength"); //signal

                QByteArray sesultStrength = replyStrength.value().toByteArray();
                int wifiSignal = sesultStrength[0] & 0x000000FF;
                wifiSignal |= ((sesultStrength[1] << 8) & 0x0000FF00);
                wifiSignal |= ((sesultStrength[2] << 16) & 0x00FF0000);
                wifiSignal |= ((sesultStrength[3] << 24) & 0xFF000000);
                setNetIcon(wifiSignal, true);
            } else {
                setNetIcon(-2, true);
            }
        }
    }
    if (netconnectType == 0) {
        setNetIcon(-1, false);
    }
}

void TrayIconWatcher::setVolumeIcon(int value)
{
    if (m_type != TrayIconWatcher::Volume) {
        return;
    }
    if (value <= 0) {
        setIconSource("audio-volume-muted-symbolic");
    } else if (value > 0 && value <= 33) {
        setIconSource("audio-volume-low-symbolic");
    } else if (value > 33 && value <= 66) {
        setIconSource("audio-volume-medium-symbolic");
    } else {
        setIconSource("audio-volume-high-symbolic");
    }
}

void TrayIconWatcher::setPowerIcon(int powerStateValue, int powerValue)
{
    if (m_type != TrayIconWatcher::Power) {
        return;
    }
    if (powerValue == 100) {
        if (powerStateValue == 2 || powerStateValue == 3) {
            setIconSource("battery-level-100-symbolic");
        } else {
            setIconSource("battery-level-100-charging-symbolic");
        }
        return;
    }
    int val = 0;
    for (int i = 10; i <= 100; i += 10) {
        if (powerValue <= i) {
            val = i - 10;
            break;
        }
    }
    QString valueString = "";
    if (powerStateValue == 2 || powerStateValue == 3) {
        valueString = QString("battery-level-%1-symbolic").arg(val);
    } else {
        valueString = QString("battery-level-%1-charging-symbolic").arg(val);
    }
    setIconSource(valueString);
}

void TrayIconWatcher::setMessageIcon(bool value)
{
    if (m_type != TrayIconWatcher::Message) {
        return;
    }
    if (value) {
        setIconSource("kylin-tool-box-null");
    } else {
        setIconSource("kylin-tool-box");
    }
}

void TrayIconWatcher::setNetIcon(int value, bool state)
{
    if (m_type != TrayIconWatcher::Network) {
        return;
    }
    if (!state) {
        setIconSource("network-wired-disconnected-symbolic");
    }

    if (netconnectType == 1) {
        if (value == -1) {
            setIconSource("network-wired-symbolic");
            return;
        }
        setIconSource("network-error-symbolic");
        return;
    }

    if (netconnectType == 2) {
        if (value < 0) {
            setIconSource("network-wired-symbolic");
            return;
        }
        //无线网络状态改变
        if (value >= 75) {
            setIconSource("network-wireless-signal-excellent-symbolic");
        } else if (value >= 55 && value < 75) {
            setIconSource("network-wireless-signal-good-symbolic");
        } else if (value >= 35 && value < 55) {
            setIconSource("network-wireless-signal-ok-symbolic");
        } else if(value >= 15 && value < 35) {
            setIconSource("network-wireless-signal-weak-symbolic");
        } else {
            setIconSource("network-wired-disconnected-symbolic");//network-cellular-signal-none-symbolic
        }
    }
}

void TrayIconWatcher::setBluetoothIcon(bool state)
{
    if (m_type != TrayIconWatcher::Bluetooth) {
        return;
    }

    if (state) {
        setIconSource("bluetooth-active-symbolic");
    } else {
        setIconSource("bluetooth-disabled-symbolic");
    }
}

void TrayIconWatcher::setVolume(int sound)
{
    setVolumeIcon(sound);
}

void TrayIconWatcher::setMuteState(bool state)
{
    //返回true表示是静音状态
    if (state) {
        setVolumeIcon(0);
    } else {
        getVolume();
    }
}
