#include "taskbarmanager.h"
#include <QDateTime>
#include <QPixmap>
#include <QFile>
#include <kdesktopfile.h>
#include <QList>
#include <QThread>

TaskBarManager::TaskBarManager(ModelManager *parent)
    : QObject(parent)
    , m_taskbarModelManager(parent)
    , m_appManagerInterface(new AppManagerInterface("com.kylin.AppManager",
                                                    "/com/kylin/AppManager",
                                                    QDBusConnection::sessionBus()))

{
    if (!m_appManagerInterface->isValid()) {
        qWarning() << "Connect to com.kylin.AppManager dbus service failed!"
                   << m_appManagerInterface->lastError();
    }
    if (QGuiApplication::platformName().startsWith(QLatin1String("wayland"), Qt::CaseInsensitive)) {
        m_isWayland = true;
    }
    if (m_isWayland) {
        connect(kdk::WindowManager::self(), &kdk::WindowManager::windowRemoved, this, &TaskBarManager::setAppCloseWayland);
    } else {
        connect(KWindowSystem::self(), &KWindowSystem::windowRemoved, this, &TaskBarManager::setAppClose);
    }

    QDBusConnection::sessionBus().connect(QString("com.kylin.AppManager"),
                                          QString("/com/kylin/AppManager"),
                                          QString("com.kylin.AppManager"),
                                          QString("appLaunched"), this, SLOT(windowOpenInformation(QString, qlonglong, qlonglong)));

}

TaskBarManager::~TaskBarManager()
{
}

void TaskBarManager::setWindowDesktopFileName(quint32 wid, const QString &fileName)
{
    m_isOpenWindows[fileName].push_back(wid);
}

void TaskBarManager::windowOpenInformation(QString desktopName, qlonglong pid, qlonglong wid)
{
    using namespace kdk;
    qDebug() << "window open desktop,pid,wid" << desktopName << wid << pid;
    if (desktopName.isEmpty()) {
        qWarning() << "desktop is null";
        return;
    }
    if (!desktopName.endsWith(".desktop")) {
        desktopName = desktopName + ".desktop";
    }

    int counter = 0;
    while (m_isWayland ? !WindowManager::windows().contains(wid) :
           !KWindowSystem::windows().contains(wid)) {
        if (counter > 5000) {
            qDebug() << __func__ << "Exit the cycle after exceeding the time!!!";
            break;
        }
        qApp->processEvents();
        ++counter;
        QThread::msleep(1);
    }

    bool pidExist = false;
    if (m_isWayland) {
        QList<WindowId> widList = WindowManager::windows();
        for (int i = 0; i < widList.count(); ++i) {
            if (WindowManager::getPid(widList[i]) == pid && widList.contains(wid)) {
                pidExist = true;
                break;
            }
        }
    } else {
        QList<WId> widList = KWindowSystem::windows();
        for (int i = 0; i < widList.count(); ++i) {
            KWindowInfo windowInfo(widList[i], NET::WMPid, NET::WM2DesktopFileName);
            if (windowInfo.pid() == pid && KWindowSystem::windows().contains(wid)) {
                pidExist = true;
                break;
            }
        }
    }
    if (!pidExist) {
        qDebug() << "pid is invalid" << wid << KWindowSystem::windows();
        return;
    }

    QString desktopFileName = m_taskbarModelManager->findDesktopNameFromTaskbar(desktopName);
    if (desktopFileName.isEmpty()) {
        desktopFileName = m_taskbarModelManager->findDesktopName(desktopName);
        if (desktopFileName.isEmpty()) {
            qWarning() << __func__ << "desktopFileName not found !!!";
            return;
        }
        if (m_isOpenWindows.value(desktopName).contains(wid)) {
            return;
        }
        QString iconFileName = m_taskbarModelManager->findIconPath(desktopFileName);
        QString appName = m_taskbarModelManager->findAppName(desktopFileName);
        m_isOpenWindows[desktopName].push_back(wid);
        m_taskbarModelManager->addWindowIconToTaskbar(iconFileName, desktopFileName, appName, wid);
    } else {
        // 该应用在任务栏，设置为打开状态
        m_taskbarModelManager->setAppOpenState(desktopName, wid);
        if (m_isOpenWindows.value(desktopName).contains(wid)) {
            return;
        }
        m_isOpenWindows[desktopName].push_back(wid);
    }
}

void TaskBarManager::setAppClose(WId wid)
{
    qDebug() << "app close wid:" << wid;
    for (auto widList = m_isOpenWindows.begin(); widList != m_isOpenWindows.end(); widList++) {
        for (auto j = 0; j < widList->count(); j++) {
            if (widList->at(j) != wid) {
                continue;
            }
            if (widList->count() == 1) {
                QString desktopName = widList.key();
                if (m_taskbarModelManager->findDesktopName(desktopName).isEmpty()) {
                    m_taskbarModelManager->setAppCloseState(desktopName);
                } else {
                    m_taskbarModelManager->removeAppIconFromTaskBar(desktopName);
                }
                m_isOpenWindows.erase(widList);
                return;
            } else {
                qDebug() << "remove window" << widList.value();
                widList->removeAt(j);
                QString desktopName = widList.key();
                if (m_taskbarModelManager->findDesktopName(desktopName).isEmpty() &&
                        m_taskbarModelManager->findDesktopNameFromTaskbar(desktopName).isEmpty()) {
                    qWarning() << __func__ << "desktopname is not found";
                    return;
                }
                m_taskbarModelManager->changeAppWindowWid(desktopName, wid, widList->at(0));
                return;
            }
        }
    }
}

void TaskBarManager::setAppCloseWayland(kdk::WindowId wid)
{
    qDebug() << "app close wid:" << wid.toUInt();
    for (auto widList = m_isOpenWindows.begin(); widList != m_isOpenWindows.end(); widList++) {
        for (auto j = 0; j < widList->count(); j++) {
            if (widList->at(j) != wid) {
                continue;
            }
            if (widList->count() == 1) {
                QString desktopName = widList.key();
                if (m_taskbarModelManager->findDesktopName(desktopName).isEmpty()) {
                    m_taskbarModelManager->setAppCloseState(desktopName);
                } else {
                    m_taskbarModelManager->removeAppIconFromTaskBar(desktopName);
                }
                m_isOpenWindows.erase(widList);
                return;
            } else {
                qDebug() << "remove window" <<widList->at(j);
                widList->removeAt(j);
                QString desktopName = widList.key();
                if (m_taskbarModelManager->findDesktopName(desktopName).isEmpty() &&
                        m_taskbarModelManager->findDesktopNameFromTaskbar(desktopName).isEmpty()) {
                    qWarning() << __func__ << "desktopname is not found";
                    return;
                }
                m_taskbarModelManager->changeAppWindowWid(desktopName, wid.toUInt(), widList->at(0));
                return;
            }
        }
    }
}
